# 49 "cil/src/frontc/clexer.mll"
 
open Cparser
exception InternalError of string
module H = Hashtbl
module E = Errorloc

let matchingParsOpen = ref 0

let currentLoc () = Cabshelper.currentLoc ()

let one_line_ghost = ref false
let is_oneline_ghost () = !one_line_ghost
let enter_oneline_ghost () = one_line_ghost := true
let exit_oneline_ghost () = one_line_ghost := false

let ghost_code = ref false
let is_ghost_code () = !ghost_code
let enter_ghost_code () = ghost_code := true
let exit_ghost_code () = ghost_code := false

let addComment c = Cabshelper.Comments.add (currentLoc()) c

(* Some debugging support for line numbers *)
let dbgToken (t: token) =
  if false then begin
    let dprintf fmt = Kernel.debug fmt in
    (match t with
         IDENT n -> dprintf "IDENT(%s)\n" n
       | LBRACE l -> dprintf "LBRACE(%d)\n" (fst l).Lexing.pos_lnum
       | RBRACE l -> dprintf "RBRACE(%d)\n" (fst l).Lexing.pos_lnum
       | IF l -> dprintf "IF(%d)\n" (fst l).Lexing.pos_lnum
       | SWITCH l -> dprintf "SWITCH(%d)\n" (fst l).Lexing.pos_lnum
       | RETURN l -> dprintf "RETURN(%d)\n" (fst l).Lexing.pos_lnum
       | _ -> ()) ;
    t
  end else
    t

(*
** Keyword hashtable
*)
let lexicon = H.create 211
let init_lexicon _ =
  H.clear lexicon;
  Logic_env.reset_typenames ();
  Logic_env.builtin_types_as_typenames ();
  List.iter
    (fun (key, builder) -> H.add lexicon key builder)
    [ ("auto", fun loc -> AUTO loc);
      ("const", fun loc -> CONST loc);
      ("__const", fun loc -> CONST loc);
      ("__const__", fun loc -> CONST loc);
      ("static", fun loc -> STATIC loc);
      ("extern", fun loc -> EXTERN loc);
      ("long", fun loc -> LONG loc);
      ("short", fun loc -> SHORT loc);
      ("register", fun loc -> REGISTER loc);
      ("signed", fun loc -> SIGNED loc);
      ("__signed", fun loc -> SIGNED loc);
      ("unsigned", fun loc -> UNSIGNED loc);
      ("volatile", fun loc -> VOLATILE loc);
      ("__volatile", fun loc -> VOLATILE loc);
      (* WW: see /usr/include/sys/cdefs.h for why __signed and __volatile
       * are accepted GCC-isms *)
      ("char", fun loc -> CHAR loc);
      ("_Bool", fun loc -> BOOL loc);
      ("int", fun loc -> INT loc);
      ("float", fun loc -> FLOAT loc);
      ("double", fun loc -> DOUBLE loc);
      ("void", fun loc -> VOID loc);
      ("enum", fun loc -> ENUM loc);
      ("struct", fun loc -> STRUCT loc);
      ("typedef", fun loc -> TYPEDEF loc);
      ("union", fun loc -> UNION loc);
      ("break", fun loc -> BREAK loc);
      ("continue", fun loc -> CONTINUE loc);
      ("goto", fun loc -> GOTO loc);
      ("return", fun loc -> dbgToken (RETURN loc));
      ("switch", fun loc -> dbgToken (SWITCH loc));
      ("case", fun loc -> CASE loc);
      ("default", fun loc -> DEFAULT loc);
      ("while", fun loc -> WHILE loc);
      ("do", fun loc -> DO loc);
      ("for", fun loc -> FOR loc);
      ("if", fun loc -> dbgToken (IF loc));
      ("else", fun _ -> ELSE);
      (*** Implementation specific keywords ***)
      ("__signed__", fun loc -> SIGNED loc);
      ("__inline__", fun loc -> INLINE loc);
      ("inline", fun loc -> INLINE loc);
      ("__inline", fun loc -> INLINE loc);
      ("_inline", fun loc ->
                      if !Cprint.msvcMode then
                        INLINE loc
                      else
                        IDENT ("_inline"));
      ("__attribute__", fun loc -> ATTRIBUTE loc);
      ("__attribute", fun loc -> ATTRIBUTE loc);
(*
      ("__attribute_used__", fun loc -> ATTRIBUTE_USED loc);
*)
      ("__blockattribute__", fun _ -> BLOCKATTRIBUTE);
      ("__blockattribute", fun _ -> BLOCKATTRIBUTE);
      ("__asm__", fun loc -> ASM loc);
      ("asm", fun loc -> ASM loc);
      ("__typeof__", fun loc -> TYPEOF loc);
      ("__typeof", fun loc -> TYPEOF loc);
      ("typeof", fun loc -> TYPEOF loc);
      ("__alignof", fun loc -> ALIGNOF loc);
      ("__alignof__", fun loc -> ALIGNOF loc);
      ("__volatile__", fun loc -> VOLATILE loc);
      ("__volatile", fun loc -> VOLATILE loc);

      ("__FUNCTION__", fun loc -> FUNCTION__ loc);
      ("__func__", fun loc -> FUNCTION__ loc); (* ISO 6.4.2.2 *)
      ("__PRETTY_FUNCTION__", fun loc -> PRETTY_FUNCTION__ loc);
      ("__label__", fun _ -> LABEL__);
      (*** weimer: GCC arcana ***)
      ("__restrict", fun loc -> RESTRICT loc);
      ("restrict", fun loc -> RESTRICT loc);
(*      ("__extension__", EXTENSION); *)
      (**** MS VC ***)
      ("__int64", fun _ -> INT64 (currentLoc ()));
      ("__int32", fun loc -> INT loc);
      ("_cdecl",  fun _ -> MSATTR ("_cdecl", currentLoc ()));
      ("__cdecl", fun _ -> MSATTR ("__cdecl", currentLoc ()));
      ("_stdcall", fun _ -> MSATTR ("_stdcall", currentLoc ()));
      ("__stdcall", fun _ -> MSATTR ("__stdcall", currentLoc ()));
      ("_fastcall", fun _ -> MSATTR ("_fastcall", currentLoc ()));
      ("__fastcall", fun _ -> MSATTR ("__fastcall", currentLoc ()));
      ("__w64", fun _ -> MSATTR("__w64", currentLoc ()));
      ("__declspec", fun loc -> DECLSPEC loc);
      ("__forceinline", fun loc -> INLINE loc); (* !! we turn forceinline
                                                 * into inline *)
      ("__try", fun loc -> TRY loc);
      ("__except", fun loc -> EXCEPT loc);
      ("__finally", fun loc -> FINALLY loc);
      (* weimer: some files produced by 'GCC -E' expect this type to be
       * defined *)
      ("__builtin_va_list",
       fun _ -> NAMED_TYPE "__builtin_va_list");
      ("__builtin_va_arg", fun loc -> BUILTIN_VA_ARG loc);
      ("__builtin_types_compatible_p", fun loc -> BUILTIN_TYPES_COMPAT loc);
      ("__builtin_offsetof", fun loc -> BUILTIN_OFFSETOF loc);
      (* On some versions of GCC __thread is a regular identifier *)
      ("__thread",
       (fun loc ->
          if Cil.theMachine.Cil.theMachine.Cil_types.__thread_is_keyword then
            THREAD loc
          else
            IDENT "__thread"));
    ]


let is_c_keyword s = Hashtbl.mem lexicon s

(* Mark an identifier as a type name. The old mapping is preserved and will
 * be reinstated when we exit this context *)
let add_type name =
   (* ignore (print_string ("adding type name " ^ name ^ "\n"));  *)
  H.add lexicon name (fun _ -> NAMED_TYPE name);
  Logic_env.add_typename name

let context : string list list ref = ref [ [] ]

let push_context _ = context := []::!context

let pop_context _ =
  match !context with
    [] -> Kernel.fatal "Empty context stack"
  | con::sub ->
		(context := sub;
		List.iter (fun name ->
                            (* Format.eprintf
                              "removing lexicon for %s@." name; *)
                             H.remove lexicon name;
                             Logic_env.remove_typename name
                          ) con)

(* Mark an identifier as a variable name. The old mapping is preserved and
 * will be reinstated when we exit this context  *)
let add_identifier name =
  match !context with
    [] -> Kernel.fatal "Empty context stack"
  | con::sub ->
      (context := (name::con)::sub;
       (*Format.eprintf "adding IDENT for %s@." name;*)
       H.add lexicon name (fun _ -> dbgToken (IDENT name));
       Logic_env.hide_typename name
      )


(*
** Useful primitives
*)
let scan_ident id =
  let here = currentLoc () in
  try (H.find lexicon id) here
  (* default to variable name, as opposed to type *)
  with Not_found -> dbgToken (IDENT id)


(*
** Buffer processor
*)


let init ~(filename: string) : Lexing.lexbuf =
  init_lexicon ();
  (* Inititialize the pointer in Errormsg *)
  Lexerhack.add_type := add_type;
  Lexerhack.push_context := push_context;
  Lexerhack.pop_context := pop_context;
  Lexerhack.add_identifier := add_identifier;
  E.startParsing ~useBasename:false filename


let finish () =
  E.finishParsing ()

(*** Error handling ***)
let error msg =
  E.parse_error msg


(*** escape character management ***)
let scan_escape (char: char) : int64 =
  let result = match char with
    'n' -> '\n'
  | 'r' -> '\r'
  | 't' -> '\t'
  | 'b' -> '\b'
  | 'f' -> '\012'  (* ASCII code 12 *)
  | 'v' -> '\011'  (* ASCII code 11 *)
  | 'a' -> '\007'  (* ASCII code 7 *)
  | 'e' | 'E' -> '\027'  (* ASCII code 27. This is a GCC extension *)
  | '\'' -> '\''
  | '"'-> '"'     (* '"' *)
  | '?' -> '?'
  | '(' when not !Cprint.msvcMode -> '('
  | '{' when not !Cprint.msvcMode -> '{'
  | '[' when not !Cprint.msvcMode -> '['
  | '%' when not !Cprint.msvcMode -> '%'
  | '\\' -> '\\'
  | other -> error ("Unrecognized escape sequence: \\" ^ (String.make 1 other))
  in
  Int64.of_int (Char.code result)

let scan_hex_escape str =
  let radix = Int64.of_int 16 in
  let the_value = ref Int64.zero in
  (* start at character 2 to skip the \x *)
  for i = 2 to (String.length str) - 1 do
    let thisDigit = Cabshelper.valueOfDigit (String.get str i) in
    (* the_value := !the_value * 16 + thisDigit *)
    the_value := Int64.add (Int64.mul !the_value radix) thisDigit
  done;
  !the_value

let scan_oct_escape str =
  let radix = Int64.of_int 8 in
  let the_value = ref Int64.zero in
  (* start at character 1 to skip the \x *)
  for i = 1 to (String.length str) - 1 do
    let thisDigit = Cabshelper.valueOfDigit (String.get str i) in
    (* the_value := !the_value * 8 + thisDigit *)
    the_value := Int64.add (Int64.mul !the_value radix) thisDigit
  done;
  !the_value

let lex_hex_escape remainder lexbuf =
  let prefix = scan_hex_escape (Lexing.lexeme lexbuf) in
  prefix :: remainder lexbuf

let lex_oct_escape remainder lexbuf =
  let prefix = scan_oct_escape (Lexing.lexeme lexbuf) in
  prefix :: remainder lexbuf

let lex_simple_escape remainder lexbuf =
  let lexchar = Lexing.lexeme_char lexbuf 1 in
  let prefix = scan_escape lexchar in
  prefix :: remainder lexbuf

let lex_unescaped remainder lexbuf =
  let prefix = Int64.of_int (Char.code (Lexing.lexeme_char lexbuf 0)) in
  prefix :: remainder lexbuf

let lex_comment remainder buffer lexbuf =
  let ch = Lexing.lexeme_char lexbuf 0 in
  if ch = '\n' then E.newline() ;
  (match buffer with None -> () | Some b -> Buffer.add_char b ch) ;
  remainder buffer lexbuf

let do_lex_comment ?first_char remainder lexbuf =
  let buffer =
    if Kernel.PrintComments.get () then
      Some(let b = Buffer.create 80 in
           (match first_char with Some c ->
             Buffer.add_char b c
           | None -> ());
           b)
    else None
  in remainder buffer lexbuf ;
  match buffer with
    | Some b -> addComment (Buffer.contents b)
    | None -> ()


(* ISO standard locale-specific function to convert a wide character
 * into a sequence of normal characters. Here we work on strings.
 * We convert L"Hi" to "H\000i\000"
  matth: this seems unused.
let wbtowc wstr =
  let len = String.length wstr in
  let dest = String.make (len * 2) '\000' in
  for i = 0 to len-1 do
    dest.[i*2] <- wstr.[i] ;
  done ;
  dest
*)

(* This function converst the "Hi" in L"Hi" to { L'H', L'i', L'\0' }
  matth: this seems unused.
let wstr_to_warray wstr =
  let len = String.length wstr in
  let res = ref "{ " in
  for i = 0 to len-1 do
    res := !res ^ (Printf.sprintf "L'%c', " wstr.[i])
  done ;
  res := !res ^ "}" ;
  !res
*)

(* Pragmas get explicit end-of-line tokens.
 * Elsewhere they are silently discarded as whitespace. *)
let pragmaLine = ref false

let annot_char = ref '@'

let () =
  Kernel.ReadAnnot.add_set_hook
    (fun _ x ->
      (* prevent the C lexer interpretation of comments *)
      annot_char := if x then '@' else '\000')

let annot_start_pos = ref Cabshelper.cabslu
let buf = Buffer.create 1024

let save_current_pos () =
  annot_start_pos := currentLoc ()

let make_annot ~one_line lexbuf s =
  let start = snd !annot_start_pos in
  let stop, token = Logic_lexer.annot (start, s) in
  lexbuf.Lexing.lex_curr_p <- stop; 
  (* The filename has already been normalized, so we must reuse it "as is". *)
  E.setCurrentFile ~normalize:false stop.Lexing.pos_fname;
  E.setCurrentLine stop.Lexing.pos_lnum;
  if one_line then E.newline ();
  match token with
    | Logic_ptree.Adecl d -> DECL d
    | Logic_ptree.Aspec -> SPEC (start,s)
        (* At this point, we only have identified a function spec. Complete
           parsing of the annotation will only occur in the cparser.mly rule.
         *)
    | Logic_ptree.Acode_annot (loc,a) -> CODE_ANNOT (a, loc)
    | Logic_ptree.Aloop_annot (loc,a) -> LOOP_ANNOT (a,loc)
    | Logic_ptree.Aattribute_annot (loc,a) -> ATTRIBUTE_ANNOT (a, loc)
    | Logic_ptree.Acustom(loc,id, a) -> CUSTOM_ANNOT(a, id, loc)


# 374 "cil/src/frontc/clexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\184\255\185\255\091\000\192\000\192\255\193\255\194\255\
    \195\255\196\255\197\255\198\255\199\255\200\255\201\255\202\255\
    \003\000\067\000\070\000\035\000\094\000\090\000\072\000\073\000\
    \091\000\172\000\074\000\170\000\013\001\051\001\244\255\124\001\
    \246\255\211\001\248\255\001\000\250\255\062\001\189\000\232\255\
    \128\000\140\000\254\255\252\255\249\255\005\000\042\002\129\002\
    \216\002\047\003\134\003\221\003\052\004\139\004\226\004\057\005\
    \144\005\231\005\062\006\149\006\236\006\067\007\154\007\241\007\
    \072\008\159\008\246\008\077\009\164\009\251\009\082\010\169\010\
    \000\011\243\255\245\255\135\000\077\011\105\011\115\011\083\000\
    \014\001\040\001\127\011\149\011\095\000\144\011\066\001\161\011\
    \183\000\065\001\085\001\140\011\102\000\184\000\141\011\086\001\
    \162\011\133\000\242\255\231\011\241\011\014\012\024\012\200\001\
    \053\012\031\002\063\012\118\002\124\012\162\012\163\000\178\000\
    \252\000\107\011\092\012\232\012\200\012\009\013\205\002\042\013\
    \036\003\240\255\249\000\156\000\126\000\183\000\049\012\183\000\
    \112\001\048\001\147\011\241\255\006\001\162\000\131\000\199\000\
    \209\000\236\255\222\255\171\000\224\000\216\000\003\000\237\255\
    \216\255\235\255\214\255\215\255\234\255\233\255\231\255\206\255\
    \230\255\207\255\229\255\228\255\221\255\016\001\227\255\220\255\
    \018\001\226\255\223\255\086\013\173\013\004\014\091\014\178\014\
    \033\001\255\255\051\001\253\255\254\255\037\001\255\255\133\001\
    \254\255\255\255\036\015\249\255\250\255\043\001\252\255\253\255\
    \166\003\255\255\251\255\041\015\249\255\229\000\243\000\051\015\
    \253\003\255\255\041\001\197\001\053\001\148\011\253\255\052\001\
    \020\001\018\001\092\001\035\001\045\001\252\255\050\001\045\001\
    \044\001\067\001\084\004\100\015\073\001\077\001\148\001\115\001\
    \103\001\143\001\157\001\169\001\171\001\205\001\003\002\177\001\
    \161\001\193\001\199\001\235\001\254\001\251\255\251\001\036\002\
    \004\002\252\001\018\002\036\002\020\002\033\002\065\002\077\002\
    \072\002\084\002\079\002\081\002\083\002\092\002\125\002\106\002\
    \121\002\124\002\109\002\158\002\163\002\165\002\176\002\176\002\
    \176\002\174\002\170\002\206\002\213\002\208\002\190\002\196\002\
    \240\002\008\003\253\002\006\003\025\003\014\003\013\003\003\003\
    \033\003\036\003\030\003\030\003\045\003\074\003\064\003\078\003\
    \060\003\075\003\074\003\075\003\066\003\077\003\093\003\081\003\
    \097\003\100\003\109\003\121\003\116\003\118\003\095\013\251\255\
    \076\015\171\004\255\255\252\255\105\015\107\015\113\015\253\255\
    \014\004\253\255\254\255\255\255\015\004\254\255\255\255\139\010\
    \250\255\251\255\148\012\255\255\252\255\046\005\124\015\179\015\
    \133\005\253\255\135\010\250\255\251\255\202\015\255\255\252\255\
    \220\005\217\015\029\016\051\006\253\255\008\013\253\255\254\255\
    \181\006\143\003\254\255\255\255\146\012\253\255\135\003\255\255\
    \161\003\144\003\151\003\254\255\082\004\253\255\254\255\157\003\
    \152\003\149\003\151\003\255\255\016\004\252\255\253\255\254\255\
    \236\003\255\255\255\004\254\255\180\003\201\003\198\003\199\003\
    \255\255\103\004\254\255\255\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\069\000\069\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \036\000\038\000\037\000\052\000\050\000\051\000\046\000\044\000\
    \043\000\042\000\047\000\064\000\017\000\017\000\255\255\069\000\
    \255\255\069\000\255\255\071\000\255\255\004\000\045\000\255\255\
    \002\000\000\000\255\255\255\255\255\255\255\255\069\000\069\000\
    \069\000\069\000\069\000\069\000\008\000\069\000\069\000\069\000\
    \069\000\066\000\069\000\069\000\069\000\069\000\067\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \068\000\255\255\255\255\255\255\255\255\255\255\013\000\255\255\
    \017\000\017\000\017\000\016\000\255\255\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\255\255\017\000\017\000\017\000\
    \017\000\255\255\255\255\255\255\013\000\255\255\013\000\255\255\
    \013\000\255\255\013\000\255\255\255\255\015\000\255\255\015\000\
    \015\000\015\000\255\255\255\255\255\255\013\000\255\255\013\000\
    \255\255\255\255\015\000\015\000\015\000\255\255\014\000\255\255\
    \014\000\014\000\014\000\255\255\014\000\014\000\014\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\030\000\255\255\255\255\
    \031\000\255\255\255\255\069\000\069\000\069\000\069\000\065\000\
    \001\000\255\255\255\255\255\255\255\255\002\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\006\000\255\255\255\255\
    \001\000\255\255\255\255\255\255\255\255\006\000\006\000\002\000\
    \001\000\255\255\255\255\002\000\002\000\002\000\255\255\002\000\
    \002\000\002\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\005\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \004\000\001\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\005\000\255\255\255\255\002\000\003\000\001\000\
    \002\000\255\255\255\255\255\255\255\255\005\000\255\255\255\255\
    \002\000\003\000\001\000\002\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\002\000\255\255\
    \255\255\255\255\255\255\255\255\003\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \003\000\255\255\002\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\255\255\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \000\000\255\255\000\000\255\255\000\000\255\255\255\255\000\000\
    \043\000\042\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\171\000\000\000\000\000\255\255\000\000\176\000\
    \000\000\000\000\179\000\000\000\000\000\255\255\000\000\000\000\
    \255\255\000\000\000\000\188\000\000\000\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\039\001\000\000\
    \044\001\255\255\000\000\000\000\044\001\044\001\044\001\000\000\
    \049\001\000\000\000\000\000\000\053\001\000\000\000\000\056\001\
    \000\000\000\000\060\001\000\000\000\000\255\255\255\255\255\255\
    \255\255\000\000\067\001\000\000\000\000\071\001\000\000\000\000\
    \255\255\255\255\255\255\255\255\000\000\078\001\000\000\000\000\
    \255\255\082\001\000\000\000\000\085\001\000\000\255\255\000\000\
    \255\255\255\255\255\255\000\000\255\255\000\000\000\000\255\255\
    \255\255\255\255\255\255\000\000\101\001\000\000\000\000\000\000\
    \255\255\000\000\255\255\000\000\255\255\255\255\255\255\255\255\
    \000\000\114\001\000\000\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\037\000\036\000\044\000\037\000\037\000\045\000\044\000\
    \000\000\000\000\045\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\037\000\000\000\000\000\000\000\000\000\000\000\
    \037\000\026\000\030\000\034\000\143\000\022\000\020\000\032\000\
    \008\000\007\000\023\000\025\000\005\000\024\000\027\000\038\000\
    \029\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\014\000\006\000\018\000\016\000\017\000\015\000\
    \162\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\031\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\010\000\035\000\009\000\019\000\033\000\
    \155\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\004\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\012\000\021\000\011\000\013\000\003\000\
    \159\000\160\000\157\000\156\000\153\000\150\000\149\000\138\000\
    \147\000\097\000\255\255\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\092\000\255\255\152\000\
    \148\000\146\000\088\000\154\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\255\255\126\000\
    \126\000\093\000\003\000\139\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\151\000\144\000\
    \136\000\125\000\100\000\100\000\100\000\100\000\100\000\100\000\
    \100\000\100\000\100\000\100\000\003\000\078\000\078\000\041\000\
    \121\000\145\000\121\000\121\000\040\000\135\000\131\000\131\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\039\000\131\000\077\000\077\000\123\000\137\000\
    \002\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\110\000\077\000\077\000\124\000\003\000\
    \140\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\163\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\078\000\078\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\037\000\
    \122\000\141\000\037\000\037\000\142\000\158\000\121\000\161\000\
    \169\000\121\000\077\000\077\000\174\000\186\000\078\000\206\000\
    \037\000\081\000\095\000\131\000\203\000\173\000\037\000\202\000\
    \198\000\078\000\080\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\028\000\028\000\077\000\121\000\078\000\
    \078\000\121\000\077\000\077\000\094\000\075\000\079\000\079\000\
    \077\000\082\000\096\000\131\000\132\000\093\000\198\000\081\000\
    \255\255\199\000\080\000\078\000\078\000\131\000\077\000\077\000\
    \080\000\198\000\198\000\076\000\255\255\077\000\089\000\177\000\
    \198\000\204\000\205\000\207\000\208\000\075\000\088\000\088\000\
    \077\000\209\000\077\000\077\000\079\000\093\000\073\000\082\000\
    \003\000\088\000\093\000\074\000\210\000\131\000\077\000\077\000\
    \080\000\198\000\198\000\076\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\088\000\088\000\
    \032\001\019\001\077\000\077\000\133\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\012\001\
    \254\000\127\000\245\000\003\000\134\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \102\000\102\000\102\000\102\000\102\000\102\000\102\000\102\000\
    \102\000\102\000\244\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\240\000\235\000\232\000\
    \231\000\200\000\230\000\225\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\047\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\194\000\226\000\
    \227\000\201\000\046\000\172\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\228\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\224\000\229\000\229\000\229\000\
    \223\000\233\000\234\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\177\000\229\000\236\000\
    \237\000\003\000\238\000\055\000\003\000\003\000\003\000\053\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\054\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\106\000\106\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \239\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\229\000\241\000\242\000\243\000\229\000\
    \229\000\246\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\247\000\248\000\249\000\250\000\
    \003\000\251\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\048\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\117\000\117\000\117\000\
    \117\000\117\000\117\000\117\000\117\000\117\000\117\000\252\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\253\000\229\000\255\000\000\001\001\001\002\001\
    \003\001\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\004\001\005\001\006\001\007\001\003\000\
    \008\001\049\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\009\001\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\010\001\011\001\229\000\013\001\014\001\015\001\016\001\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\017\001\018\001\229\000\020\001\003\000\021\001\
    \003\000\003\000\003\000\003\000\003\000\003\000\050\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\022\001\023\001\024\001\025\001\184\000\
    \026\001\027\001\184\000\184\000\028\001\029\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \184\000\030\001\031\001\229\000\033\001\034\001\184\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\035\001\036\001\037\001\229\000\003\000\088\001\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\051\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\089\001\090\001\091\001\096\001\192\000\097\001\
    \098\001\192\000\192\000\099\001\083\001\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\192\000\
    \051\001\054\001\102\001\105\001\109\001\192\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \110\001\111\001\104\001\112\001\003\000\000\000\052\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\094\001\093\001\211\000\000\000\094\001\
    \211\000\211\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\211\000\000\000\
    \000\000\115\001\094\001\000\000\211\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\255\255\
    \000\000\000\000\094\001\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\041\001\000\000\000\000\041\001\
    \041\001\095\001\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\041\001\000\000\000\000\
    \000\000\000\000\000\000\041\001\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\063\000\003\000\003\000\003\000\000\000\
    \107\001\000\000\000\000\000\000\107\001\000\000\050\001\255\255\
    \103\001\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\107\001\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\107\001\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\058\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\108\001\255\255\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\056\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\065\001\065\001\065\001\
    \065\001\065\001\065\001\065\001\065\001\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\057\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\075\001\075\001\075\001\075\001\
    \075\001\075\001\075\001\075\001\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\059\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\080\001\000\000\
    \000\000\080\001\080\001\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\080\001\
    \000\000\000\000\000\000\000\000\000\000\080\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\060\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\061\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\062\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\064\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\065\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\066\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\067\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\068\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\069\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\070\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\059\001\070\001\000\000\
    \000\000\071\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\069\001\000\000\000\000\000\000\058\001\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \072\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\108\000\000\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\068\001\
    \000\000\000\000\000\000\057\001\000\000\000\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\107\000\000\000\107\000\000\000\
    \000\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \106\000\106\000\106\000\100\000\100\000\100\000\100\000\100\000\
    \100\000\100\000\100\000\100\000\100\000\078\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\000\000\000\000\000\000\000\000\
    \099\000\098\000\078\000\078\000\000\000\000\000\078\000\098\000\
    \121\000\000\000\000\000\078\000\077\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\028\000\028\000\078\000\
    \078\000\077\000\077\000\000\000\093\000\077\000\000\000\122\000\
    \099\000\098\000\077\000\000\000\090\000\000\000\000\000\098\000\
    \121\000\086\000\093\000\000\000\077\000\000\000\077\000\077\000\
    \131\000\198\000\085\000\094\000\000\000\000\000\000\000\000\000\
    \000\000\077\000\077\000\000\000\093\000\077\000\088\000\000\000\
    \088\000\084\000\077\000\000\000\091\000\000\000\084\000\132\000\
    \199\000\087\000\093\000\000\000\000\000\000\000\077\000\077\000\
    \131\000\198\000\085\000\000\000\000\000\089\000\093\000\000\000\
    \000\000\000\000\105\000\000\000\105\000\000\000\088\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\100\000\100\000\100\000\100\000\100\000\100\000\100\000\
    \100\000\100\000\100\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\101\000\098\000\
    \000\000\103\000\000\000\103\000\000\000\098\000\102\000\102\000\
    \102\000\102\000\102\000\102\000\102\000\102\000\102\000\102\000\
    \102\000\102\000\102\000\102\000\102\000\102\000\102\000\102\000\
    \102\000\102\000\000\000\000\000\000\000\000\000\101\000\098\000\
    \000\000\000\000\000\000\000\000\000\000\098\000\098\000\000\000\
    \000\000\126\000\126\000\000\000\098\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\106\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \106\000\000\000\000\000\098\000\000\000\129\000\098\000\000\000\
    \000\000\098\000\000\000\000\000\098\000\098\000\128\000\120\000\
    \000\000\120\000\000\000\098\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\000\000\000\000\
    \000\000\000\000\127\000\098\000\087\001\130\000\000\000\000\000\
    \000\000\098\000\000\000\000\000\000\000\098\000\128\000\000\000\
    \000\000\000\000\000\000\098\000\115\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\115\000\115\000\115\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\000\000\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\000\000\087\001\000\000\000\000\
    \115\000\000\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\000\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\000\000\000\000\000\000\062\001\000\000\112\000\000\000\
    \000\000\086\001\114\000\118\000\000\000\118\000\000\000\111\000\
    \117\000\117\000\117\000\117\000\117\000\117\000\117\000\117\000\
    \117\000\117\000\000\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\000\000\000\000\110\000\062\001\000\000\113\000\087\001\
    \000\000\080\001\114\000\000\000\080\001\080\001\000\000\111\000\
    \115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\
    \115\000\115\000\080\001\000\000\000\000\000\000\000\000\000\000\
    \080\001\115\000\115\000\115\000\115\000\115\000\115\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \116\000\117\000\117\000\117\000\117\000\117\000\117\000\117\000\
    \117\000\117\000\117\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\115\000\115\000\115\000\115\000\115\000\115\000\098\000\
    \000\000\000\000\000\000\000\000\000\000\098\000\000\000\000\000\
    \116\000\000\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\000\000\000\000\000\000\000\000\
    \041\001\042\001\000\000\041\001\041\001\000\000\000\000\098\000\
    \098\000\000\000\000\000\000\000\000\000\098\000\098\000\000\000\
    \000\000\041\001\003\000\000\000\000\000\000\000\000\000\041\001\
    \000\000\040\001\000\000\079\001\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \098\000\000\000\255\255\000\000\255\255\000\000\098\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \164\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \255\255\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\165\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\255\255\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\166\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\167\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\184\000\185\000\000\000\
    \184\000\184\000\192\000\193\000\000\000\192\000\192\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\184\000\000\000\
    \000\000\000\000\000\000\192\000\184\000\000\000\180\000\000\000\
    \000\000\192\000\000\000\000\000\183\000\182\000\000\000\000\000\
    \000\000\000\000\000\000\181\000\000\000\255\255\000\000\000\000\
    \255\255\191\000\191\000\191\000\191\000\191\000\191\000\191\000\
    \191\000\191\000\191\000\191\000\191\000\191\000\191\000\191\000\
    \191\000\191\000\191\000\191\000\191\000\211\000\043\001\000\000\
    \211\000\211\000\255\255\000\000\255\255\255\255\000\000\255\255\
    \000\000\000\000\255\255\045\001\000\000\255\255\211\000\196\000\
    \000\000\000\000\000\000\000\000\211\000\000\000\000\000\000\000\
    \195\000\000\000\000\000\043\001\000\000\043\001\000\000\000\000\
    \000\000\000\000\000\000\047\001\000\000\190\000\000\000\000\000\
    \045\001\189\000\046\001\000\000\194\000\000\000\000\000\197\000\
    \046\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \195\000\000\000\000\000\221\000\063\001\063\001\063\001\063\001\
    \063\001\063\001\063\001\063\001\063\001\063\001\000\000\000\000\
    \214\000\000\000\000\000\000\000\000\000\063\001\063\001\063\001\
    \063\001\063\001\063\001\000\000\000\000\217\000\000\000\000\000\
    \000\000\000\000\000\000\213\000\000\000\220\000\000\000\000\000\
    \212\000\000\000\000\000\218\000\000\000\000\000\215\000\219\000\
    \000\000\216\000\000\000\222\000\000\000\063\001\063\001\063\001\
    \063\001\063\001\063\001\063\001\063\001\063\001\063\001\063\001\
    \063\001\063\001\063\001\063\001\063\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\063\001\063\001\063\001\063\001\
    \063\001\063\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\074\001\074\001\074\001\074\001\074\001\074\001\074\001\
    \074\001\074\001\074\001\000\000\063\001\063\001\063\001\063\001\
    \063\001\063\001\074\001\074\001\074\001\074\001\074\001\074\001\
    \000\000\000\000\073\001\000\000\255\255\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\074\001\074\001\074\001\074\001\074\001\074\001\
    \000\000\000\000\073\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\074\001\074\001\074\001\
    \074\001\074\001\074\001\074\001\074\001\074\001\074\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\074\001\074\001\
    \074\001\074\001\074\001\074\001\000\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\074\001\074\001\
    \074\001\074\001\074\001\074\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\035\000\000\000\000\000\035\000\045\000\
    \255\255\255\255\045\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\142\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \016\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \019\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \017\000\017\000\018\000\018\000\020\000\022\000\023\000\026\000\
    \024\000\079\000\040\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\084\000\041\000\021\000\
    \024\000\024\000\092\000\020\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\041\000\075\000\
    \075\000\097\000\003\000\026\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\021\000\025\000\
    \027\000\110\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\004\000\088\000\093\000\038\000\
    \123\000\025\000\124\000\125\000\038\000\127\000\133\000\134\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\038\000\135\000\088\000\093\000\111\000\136\000\
    \000\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\111\000\088\000\093\000\111\000\004\000\
    \139\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\028\000\080\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\037\000\
    \112\000\140\000\037\000\037\000\141\000\157\000\122\000\160\000\
    \168\000\112\000\028\000\080\000\173\000\181\000\081\000\189\000\
    \037\000\028\000\080\000\132\000\190\000\170\000\037\000\194\000\
    \200\000\029\000\028\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\081\000\122\000\089\000\
    \086\000\112\000\028\000\080\000\081\000\029\000\028\000\080\000\
    \029\000\028\000\080\000\132\000\129\000\081\000\201\000\029\000\
    \040\000\196\000\028\000\090\000\095\000\129\000\089\000\086\000\
    \029\000\199\000\196\000\029\000\041\000\081\000\086\000\175\000\
    \202\000\203\000\204\000\206\000\207\000\029\000\089\000\086\000\
    \029\000\208\000\090\000\095\000\029\000\081\000\031\000\029\000\
    \031\000\090\000\095\000\031\000\209\000\129\000\089\000\086\000\
    \029\000\199\000\196\000\029\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\089\000\086\000\
    \212\000\213\000\090\000\095\000\128\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\214\000\
    \215\000\128\000\216\000\031\000\128\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\033\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\217\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\218\000\219\000\220\000\
    \221\000\195\000\223\000\224\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\195\000\225\000\
    \226\000\195\000\033\000\170\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\046\000\105\000\
    \105\000\105\000\105\000\105\000\105\000\105\000\105\000\105\000\
    \105\000\227\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\222\000\228\000\230\000\231\000\
    \222\000\232\000\233\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\175\000\234\000\235\000\
    \236\000\046\000\237\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\047\000\107\000\107\000\
    \107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\
    \238\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\239\000\240\000\241\000\242\000\243\000\
    \244\000\245\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\246\000\247\000\248\000\249\000\
    \047\000\250\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\
    \047\000\047\000\047\000\047\000\048\000\118\000\118\000\118\000\
    \118\000\118\000\118\000\118\000\118\000\118\000\118\000\251\000\
    \048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\252\000\253\000\254\000\255\000\000\001\001\001\
    \002\001\048\000\048\000\048\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\048\000\003\001\004\001\005\001\006\001\048\000\
    \007\001\048\000\048\000\048\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\048\000\049\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\008\001\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\009\001\010\001\011\001\012\001\013\001\014\001\015\001\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\016\001\017\001\018\001\019\001\049\000\020\001\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\050\000\021\001\022\001\023\001\024\001\184\000\
    \025\001\026\001\184\000\184\000\027\001\028\001\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \184\000\029\001\030\001\031\001\032\001\033\001\184\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\034\001\035\001\036\001\037\001\050\000\086\001\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\051\000\088\001\089\001\090\001\095\001\192\000\096\001\
    \097\001\192\000\192\000\098\001\081\001\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\192\000\
    \048\001\052\001\100\001\104\001\108\001\192\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \109\001\110\001\100\001\111\001\051\000\255\255\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \052\000\255\255\255\255\092\001\092\001\210\000\255\255\092\001\
    \210\000\210\000\255\255\255\255\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\210\000\255\255\
    \255\255\113\001\092\001\255\255\210\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\081\001\
    \255\255\255\255\092\001\052\000\255\255\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\053\000\
    \255\255\255\255\255\255\255\255\041\001\255\255\255\255\041\001\
    \041\001\092\001\255\255\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\041\001\255\255\255\255\
    \255\255\255\255\255\255\041\001\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\255\255\255\255\
    \255\255\255\255\053\000\255\255\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\054\000\255\255\
    \106\001\255\255\255\255\255\255\106\001\255\255\048\001\052\001\
    \100\001\255\255\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\255\255\255\255\255\255\106\001\
    \255\255\255\255\255\255\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\255\255\255\255\106\001\
    \255\255\054\000\255\255\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\055\000\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\106\001\113\001\
    \255\255\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\255\255\255\255\255\255\255\255\
    \055\000\255\255\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\056\000\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\255\255\255\255\255\255\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\255\255\255\255\255\255\255\255\056\000\
    \255\255\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\057\000\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\255\255\255\255\255\255\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\255\255\255\255\255\255\255\255\057\000\255\255\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\058\000\075\001\075\001\075\001\075\001\075\001\
    \075\001\075\001\075\001\255\255\255\255\255\255\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\255\255\255\255\255\255\255\255\058\000\255\255\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\059\000\255\255\255\255\255\255\255\255\080\001\255\255\
    \255\255\080\001\080\001\255\255\255\255\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\080\001\
    \255\255\255\255\255\255\255\255\255\255\080\001\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \255\255\255\255\255\255\255\255\059\000\255\255\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \060\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\255\255\
    \255\255\255\255\255\255\060\000\255\255\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\061\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\255\255\255\255\
    \255\255\255\255\061\000\255\255\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\062\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\255\255\255\255\255\255\
    \255\255\062\000\255\255\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\063\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\255\255\255\255\255\255\255\255\
    \063\000\255\255\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\064\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\255\255\255\255\255\255\255\255\064\000\
    \255\255\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\065\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\255\255\255\255\255\255\255\255\065\000\255\255\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\066\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\255\255\255\255\255\255\255\255\066\000\255\255\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\067\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \255\255\255\255\255\255\255\255\067\000\255\255\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \068\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\255\255\
    \255\255\255\255\255\255\068\000\255\255\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\069\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\255\255\255\255\
    \255\255\255\255\069\000\255\255\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\070\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\055\001\066\001\255\255\
    \255\255\070\000\255\255\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\071\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\066\001\255\255\255\255\255\255\055\001\
    \255\255\255\255\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\255\255\255\255\255\255\255\255\
    \071\000\255\255\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\072\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\255\255\255\255\255\255\255\255\072\000\
    \255\255\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\076\000\255\255\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\066\001\
    \255\255\255\255\255\255\055\001\255\255\255\255\076\000\076\000\
    \076\000\076\000\076\000\076\000\077\000\255\255\077\000\255\255\
    \255\255\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\082\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\255\255\255\255\255\255\255\255\
    \078\000\078\000\091\000\094\000\255\255\255\255\085\000\078\000\
    \113\000\255\255\255\255\083\000\082\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\087\000\
    \096\000\091\000\094\000\255\255\082\000\085\000\255\255\113\000\
    \078\000\078\000\083\000\255\255\085\000\255\255\255\255\078\000\
    \113\000\083\000\094\000\255\255\082\000\255\255\087\000\096\000\
    \130\000\197\000\083\000\082\000\255\255\255\255\255\255\255\255\
    \255\255\091\000\094\000\255\255\082\000\085\000\087\000\255\255\
    \091\000\085\000\083\000\255\255\085\000\255\255\083\000\130\000\
    \197\000\083\000\094\000\255\255\255\255\255\255\087\000\096\000\
    \130\000\197\000\083\000\255\255\255\255\087\000\096\000\255\255\
    \255\255\255\255\099\000\255\255\099\000\255\255\087\000\099\000\
    \099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\
    \099\000\100\000\100\000\100\000\100\000\100\000\100\000\100\000\
    \100\000\100\000\100\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\100\000\100\000\
    \255\255\101\000\255\255\101\000\255\255\100\000\101\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \102\000\102\000\102\000\102\000\102\000\102\000\102\000\102\000\
    \102\000\102\000\255\255\255\255\255\255\255\255\100\000\100\000\
    \255\255\255\255\255\255\255\255\255\255\100\000\102\000\255\255\
    \255\255\126\000\126\000\255\255\102\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\106\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \106\000\255\255\255\255\104\000\255\255\126\000\102\000\255\255\
    \255\255\104\000\255\255\255\255\102\000\106\000\126\000\114\000\
    \255\255\114\000\255\255\106\000\114\000\114\000\114\000\114\000\
    \114\000\114\000\114\000\114\000\114\000\114\000\255\255\255\255\
    \255\255\255\255\126\000\104\000\084\001\126\000\255\255\255\255\
    \255\255\104\000\255\255\255\255\255\255\106\000\126\000\255\255\
    \255\255\255\255\255\255\106\000\108\000\108\000\108\000\108\000\
    \108\000\108\000\108\000\108\000\108\000\108\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\108\000\108\000\108\000\
    \108\000\108\000\108\000\255\255\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\255\255\084\001\255\255\255\255\
    \109\000\255\255\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\255\255\108\000\108\000\108\000\
    \108\000\108\000\108\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\255\255\255\255\255\255\058\001\255\255\109\000\255\255\
    \255\255\084\001\109\000\116\000\255\255\116\000\255\255\109\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\116\000\116\000\
    \116\000\116\000\255\255\109\000\109\000\109\000\109\000\109\000\
    \109\000\255\255\255\255\109\000\058\001\255\255\109\000\084\001\
    \255\255\077\001\109\000\255\255\077\001\077\001\255\255\109\000\
    \115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\
    \115\000\115\000\077\001\255\255\255\255\255\255\255\255\255\255\
    \077\001\115\000\115\000\115\000\115\000\115\000\115\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \115\000\117\000\117\000\117\000\117\000\117\000\117\000\117\000\
    \117\000\117\000\117\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\115\000\115\000\115\000\115\000\115\000\115\000\117\000\
    \255\255\255\255\255\255\255\255\255\255\117\000\255\255\255\255\
    \115\000\255\255\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\255\255\255\255\255\255\255\255\
    \038\001\038\001\255\255\038\001\038\001\255\255\255\255\117\000\
    \119\000\255\255\255\255\255\255\255\255\117\000\119\000\255\255\
    \255\255\038\001\163\000\255\255\255\255\255\255\255\255\038\001\
    \255\255\038\001\255\255\077\001\255\255\255\255\163\000\163\000\
    \163\000\163\000\163\000\163\000\163\000\163\000\163\000\163\000\
    \119\000\255\255\084\001\255\255\058\001\255\255\119\000\163\000\
    \163\000\163\000\163\000\163\000\163\000\163\000\163\000\163\000\
    \163\000\163\000\163\000\163\000\163\000\163\000\163\000\163\000\
    \163\000\163\000\163\000\163\000\163\000\163\000\163\000\163\000\
    \163\000\255\255\255\255\255\255\255\255\163\000\255\255\163\000\
    \163\000\163\000\163\000\163\000\163\000\163\000\163\000\163\000\
    \163\000\163\000\163\000\163\000\163\000\163\000\163\000\163\000\
    \163\000\163\000\163\000\163\000\163\000\163\000\163\000\163\000\
    \163\000\164\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\164\000\164\000\164\000\
    \164\000\164\000\164\000\164\000\164\000\164\000\164\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\164\000\164\000\
    \164\000\164\000\164\000\164\000\164\000\164\000\164\000\164\000\
    \164\000\164\000\164\000\164\000\164\000\164\000\164\000\164\000\
    \164\000\164\000\164\000\164\000\164\000\164\000\164\000\164\000\
    \077\001\255\255\255\255\255\255\164\000\255\255\164\000\164\000\
    \164\000\164\000\164\000\164\000\164\000\164\000\164\000\164\000\
    \164\000\164\000\164\000\164\000\164\000\164\000\164\000\164\000\
    \164\000\164\000\164\000\164\000\164\000\164\000\164\000\164\000\
    \165\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\038\001\
    \255\255\255\255\255\255\165\000\255\255\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
    \165\000\165\000\165\000\165\000\165\000\165\000\165\000\166\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\166\000\166\000\166\000\166\000\166\000\
    \166\000\166\000\166\000\166\000\166\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\166\000\166\000\166\000\166\000\
    \166\000\166\000\166\000\166\000\166\000\166\000\166\000\166\000\
    \166\000\166\000\166\000\166\000\166\000\166\000\166\000\166\000\
    \166\000\166\000\166\000\166\000\166\000\166\000\255\255\255\255\
    \255\255\255\255\166\000\255\255\166\000\166\000\166\000\166\000\
    \166\000\166\000\166\000\166\000\166\000\166\000\166\000\166\000\
    \166\000\166\000\166\000\166\000\166\000\166\000\166\000\166\000\
    \166\000\166\000\166\000\166\000\166\000\166\000\167\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\255\255\255\255\255\255\
    \255\255\167\000\255\255\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\178\000\178\000\255\255\
    \178\000\178\000\187\000\187\000\255\255\187\000\187\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\178\000\255\255\
    \255\255\255\255\255\255\187\000\178\000\255\255\178\000\255\255\
    \255\255\187\000\255\255\255\255\178\000\178\000\255\255\255\255\
    \255\255\255\255\255\255\178\000\255\255\040\001\255\255\255\255\
    \040\001\187\000\187\000\187\000\187\000\187\000\187\000\187\000\
    \187\000\187\000\187\000\191\000\191\000\191\000\191\000\191\000\
    \191\000\191\000\191\000\191\000\191\000\211\000\040\001\255\255\
    \211\000\211\000\044\001\255\255\045\001\044\001\255\255\045\001\
    \255\255\255\255\046\001\040\001\255\255\046\001\211\000\191\000\
    \255\255\255\255\255\255\255\255\211\000\255\255\255\255\255\255\
    \191\000\255\255\255\255\044\001\255\255\045\001\255\255\255\255\
    \255\255\255\255\255\255\046\001\255\255\187\000\255\255\255\255\
    \044\001\187\000\045\001\255\255\191\000\255\255\255\255\191\000\
    \046\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \191\000\255\255\255\255\211\000\062\001\062\001\062\001\062\001\
    \062\001\062\001\062\001\062\001\062\001\062\001\255\255\255\255\
    \211\000\255\255\255\255\255\255\255\255\062\001\062\001\062\001\
    \062\001\062\001\062\001\255\255\255\255\211\000\255\255\255\255\
    \255\255\255\255\255\255\211\000\255\255\211\000\255\255\255\255\
    \211\000\255\255\255\255\211\000\255\255\255\255\211\000\211\000\
    \255\255\211\000\255\255\211\000\255\255\062\001\062\001\062\001\
    \062\001\062\001\062\001\063\001\063\001\063\001\063\001\063\001\
    \063\001\063\001\063\001\063\001\063\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\063\001\063\001\063\001\063\001\
    \063\001\063\001\069\001\069\001\069\001\069\001\069\001\069\001\
    \069\001\069\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\255\255\063\001\063\001\063\001\063\001\
    \063\001\063\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \255\255\255\255\069\001\255\255\178\000\255\255\255\255\255\255\
    \255\255\187\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\073\001\073\001\073\001\073\001\073\001\073\001\
    \255\255\255\255\069\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\040\001\074\001\074\001\074\001\
    \074\001\074\001\074\001\074\001\074\001\074\001\074\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\074\001\074\001\
    \074\001\074\001\074\001\074\001\255\255\255\255\255\255\255\255\
    \255\255\044\001\255\255\045\001\255\255\255\255\255\255\255\255\
    \255\255\046\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\074\001\074\001\
    \074\001\074\001\074\001\074\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\069\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\002\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\004\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\001\000\001\000\001\000\001\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\001\000\000\000\001\000\000\000\000\000\000\000\
    \001\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\210\000\255\255\211\000\210\000\210\000\211\000\211\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\210\000\255\255\211\000\255\255\255\255\255\255\
    \210\000\255\255\211\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_code = 
   "\255\001\255\255\000\001\255";
}

let rec initial lexbuf =
    __ocaml_lex_initial_rec lexbuf 0
and __ocaml_lex_initial_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 478 "cil/src/frontc/clexer.mll"
      (
	do_lex_comment comment lexbuf ;
        initial lexbuf
      )
# 1844 "cil/src/frontc/clexer.ml"

  | 1 ->
let
# 483 "cil/src/frontc/clexer.mll"
                        c
# 1850 "cil/src/frontc/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 484 "cil/src/frontc/clexer.mll"
    ( if c = !annot_char then begin
	Cabshelper.continue_annot
	  (currentLoc ())
	  (fun () ->
             save_current_pos ();
	     Buffer.clear buf;
	     annot_first_token lexbuf)
	  (fun () ->
	     initial lexbuf)
	  "Skipping annotation"
      end else
	begin
	  do_lex_comment ~first_char:c comment lexbuf ;
          initial lexbuf
	end
    )
# 1869 "cil/src/frontc/clexer.ml"

  | 2 ->
# 502 "cil/src/frontc/clexer.mll"
    ( do_lex_comment onelinecomment lexbuf ;
      E.newline();
      if is_oneline_ghost () then begin
        exit_oneline_ghost ();
        RGHOST
      end else begin
        initial lexbuf
      end
    )
# 1882 "cil/src/frontc/clexer.ml"

  | 3 ->
let
# 512 "cil/src/frontc/clexer.mll"
                    c
# 1888 "cil/src/frontc/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 513 "cil/src/frontc/clexer.mll"
    ( if c = !annot_char then begin
	Cabshelper.continue_annot
	  (currentLoc())
	  (fun () ->
             save_current_pos ();
	     Buffer.clear buf;
	     annot_one_line lexbuf)
	  (fun () -> initial lexbuf)
	  "Skipping annotation"
      end else
	begin
	  do_lex_comment ~first_char:c onelinecomment lexbuf ;
	  E.newline();
	  if is_oneline_ghost () then
	    begin
              exit_oneline_ghost ();
              RGHOST
	    end
	  else
	    begin
              initial lexbuf
	    end
	end
    )
# 1915 "cil/src/frontc/clexer.ml"

  | 4 ->
# 537 "cil/src/frontc/clexer.mll"
           (initial lexbuf)
# 1920 "cil/src/frontc/clexer.ml"

  | 5 ->
# 538 "cil/src/frontc/clexer.mll"
                                        ( E.newline ();
                                          if !pragmaLine then
                                            begin
                                              pragmaLine := false;
                                              PRAGMA_EOL
                                            end
                                          else if is_oneline_ghost () then
                                            begin
                                              exit_oneline_ghost ();
                                              RGHOST
                                            end
                                          else begin
                                            initial lexbuf
                                          end )
# 1938 "cil/src/frontc/clexer.ml"

  | 6 ->
# 552 "cil/src/frontc/clexer.mll"
                                        ( E.newline ();
                                          initial lexbuf
                                        )
# 1945 "cil/src/frontc/clexer.ml"

  | 7 ->
# 555 "cil/src/frontc/clexer.mll"
         ( hash lexbuf)
# 1950 "cil/src/frontc/clexer.ml"

  | 8 ->
# 556 "cil/src/frontc/clexer.mll"
                                   ( PRAGMA (currentLoc ()) )
# 1955 "cil/src/frontc/clexer.ml"

  | 9 ->
# 557 "cil/src/frontc/clexer.mll"
          ( CST_CHAR (chr lexbuf, currentLoc ()))
# 1960 "cil/src/frontc/clexer.ml"

  | 10 ->
# 558 "cil/src/frontc/clexer.mll"
          ( CST_WCHAR (chr lexbuf, currentLoc ()) )
# 1965 "cil/src/frontc/clexer.ml"

  | 11 ->
# 559 "cil/src/frontc/clexer.mll"
         (  
(* matth: BUG:  this could be either a regular string or a wide string.
 *  e.g. if it's the "world" in
 *     L"Hello, " "world"
 *  then it should be treated as wide even though there's no L immediately
 *  preceding it.  See test/small1/wchar5.c for a failure case. *)
                                          try CST_STRING (str lexbuf, currentLoc ())
                                          with e ->
                                             raise (InternalError
                                                     ("str: " ^
                                                      Printexc.to_string e)))
# 1980 "cil/src/frontc/clexer.ml"

  | 12 ->
# 570 "cil/src/frontc/clexer.mll"
           ( (* weimer: wchar_t string literal *)
                                          try CST_WSTRING(str lexbuf, currentLoc ())
                                          with e ->
                                             raise (InternalError
                                                     ("wide string: " ^
                                                      Printexc.to_string e)))
# 1990 "cil/src/frontc/clexer.ml"

  | 13 ->
# 576 "cil/src/frontc/clexer.mll"
             (CST_FLOAT (Lexing.lexeme lexbuf, currentLoc ()))
# 1995 "cil/src/frontc/clexer.ml"

  | 14 ->
# 577 "cil/src/frontc/clexer.mll"
                           ( (* GCC Extension for binary numbers *) 
                                          CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2001 "cil/src/frontc/clexer.ml"

  | 15 ->
# 579 "cil/src/frontc/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2006 "cil/src/frontc/clexer.ml"

  | 16 ->
# 580 "cil/src/frontc/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2011 "cil/src/frontc/clexer.ml"

  | 17 ->
# 581 "cil/src/frontc/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2016 "cil/src/frontc/clexer.ml"

  | 18 ->
# 582 "cil/src/frontc/clexer.mll"
             (EOF)
# 2021 "cil/src/frontc/clexer.ml"

  | 19 ->
# 583 "cil/src/frontc/clexer.mll"
           (ELLIPSIS)
# 2026 "cil/src/frontc/clexer.ml"

  | 20 ->
# 584 "cil/src/frontc/clexer.mll"
          (PLUS_EQ)
# 2031 "cil/src/frontc/clexer.ml"

  | 21 ->
# 585 "cil/src/frontc/clexer.mll"
          (MINUS_EQ)
# 2036 "cil/src/frontc/clexer.ml"

  | 22 ->
# 586 "cil/src/frontc/clexer.mll"
          (STAR_EQ)
# 2041 "cil/src/frontc/clexer.ml"

  | 23 ->
# 587 "cil/src/frontc/clexer.mll"
          (SLASH_EQ)
# 2046 "cil/src/frontc/clexer.ml"

  | 24 ->
# 588 "cil/src/frontc/clexer.mll"
          (PERCENT_EQ)
# 2051 "cil/src/frontc/clexer.ml"

  | 25 ->
# 589 "cil/src/frontc/clexer.mll"
          (PIPE_EQ)
# 2056 "cil/src/frontc/clexer.ml"

  | 26 ->
# 590 "cil/src/frontc/clexer.mll"
          (AND_EQ)
# 2061 "cil/src/frontc/clexer.ml"

  | 27 ->
# 591 "cil/src/frontc/clexer.mll"
          (CIRC_EQ)
# 2066 "cil/src/frontc/clexer.ml"

  | 28 ->
# 592 "cil/src/frontc/clexer.mll"
           (INF_INF_EQ)
# 2071 "cil/src/frontc/clexer.ml"

  | 29 ->
# 593 "cil/src/frontc/clexer.mll"
           (SUP_SUP_EQ)
# 2076 "cil/src/frontc/clexer.ml"

  | 30 ->
# 594 "cil/src/frontc/clexer.mll"
          (INF_INF)
# 2081 "cil/src/frontc/clexer.ml"

  | 31 ->
# 595 "cil/src/frontc/clexer.mll"
          (SUP_SUP)
# 2086 "cil/src/frontc/clexer.ml"

  | 32 ->
# 596 "cil/src/frontc/clexer.mll"
           (EQ_EQ)
# 2091 "cil/src/frontc/clexer.ml"

  | 33 ->
# 597 "cil/src/frontc/clexer.mll"
           (EXCLAM_EQ)
# 2096 "cil/src/frontc/clexer.ml"

  | 34 ->
# 598 "cil/src/frontc/clexer.mll"
          (INF_EQ)
# 2101 "cil/src/frontc/clexer.ml"

  | 35 ->
# 599 "cil/src/frontc/clexer.mll"
          (SUP_EQ)
# 2106 "cil/src/frontc/clexer.ml"

  | 36 ->
# 600 "cil/src/frontc/clexer.mll"
         (EQ)
# 2111 "cil/src/frontc/clexer.ml"

  | 37 ->
# 601 "cil/src/frontc/clexer.mll"
         (INF)
# 2116 "cil/src/frontc/clexer.ml"

  | 38 ->
# 602 "cil/src/frontc/clexer.mll"
         (SUP)
# 2121 "cil/src/frontc/clexer.ml"

  | 39 ->
# 603 "cil/src/frontc/clexer.mll"
          (PLUS_PLUS (currentLoc ()))
# 2126 "cil/src/frontc/clexer.ml"

  | 40 ->
# 604 "cil/src/frontc/clexer.mll"
          (MINUS_MINUS (currentLoc ()))
# 2131 "cil/src/frontc/clexer.ml"

  | 41 ->
# 605 "cil/src/frontc/clexer.mll"
          (ARROW)
# 2136 "cil/src/frontc/clexer.ml"

  | 42 ->
# 606 "cil/src/frontc/clexer.mll"
         (PLUS (currentLoc ()))
# 2141 "cil/src/frontc/clexer.ml"

  | 43 ->
# 607 "cil/src/frontc/clexer.mll"
         (MINUS (currentLoc ()))
# 2146 "cil/src/frontc/clexer.ml"

  | 44 ->
# 609 "cil/src/frontc/clexer.mll"
                    ( if is_ghost_code () then might_end_ghost lexbuf
                      else
                        STAR (currentLoc ()))
# 2153 "cil/src/frontc/clexer.ml"

  | 45 ->
# 612 "cil/src/frontc/clexer.mll"
         (SLASH)
# 2158 "cil/src/frontc/clexer.ml"

  | 46 ->
# 613 "cil/src/frontc/clexer.mll"
         (PERCENT)
# 2163 "cil/src/frontc/clexer.ml"

  | 47 ->
# 614 "cil/src/frontc/clexer.mll"
         (EXCLAM (currentLoc ()))
# 2168 "cil/src/frontc/clexer.ml"

  | 48 ->
# 615 "cil/src/frontc/clexer.mll"
          (AND_AND (currentLoc ()))
# 2173 "cil/src/frontc/clexer.ml"

  | 49 ->
# 616 "cil/src/frontc/clexer.mll"
          (PIPE_PIPE)
# 2178 "cil/src/frontc/clexer.ml"

  | 50 ->
# 617 "cil/src/frontc/clexer.mll"
         (AND (currentLoc ()))
# 2183 "cil/src/frontc/clexer.ml"

  | 51 ->
# 618 "cil/src/frontc/clexer.mll"
         (PIPE)
# 2188 "cil/src/frontc/clexer.ml"

  | 52 ->
# 619 "cil/src/frontc/clexer.mll"
         (CIRC)
# 2193 "cil/src/frontc/clexer.ml"

  | 53 ->
# 620 "cil/src/frontc/clexer.mll"
         (QUEST)
# 2198 "cil/src/frontc/clexer.ml"

  | 54 ->
# 621 "cil/src/frontc/clexer.mll"
         (COLON)
# 2203 "cil/src/frontc/clexer.ml"

  | 55 ->
# 622 "cil/src/frontc/clexer.mll"
                (TILDE (currentLoc ()))
# 2208 "cil/src/frontc/clexer.ml"

  | 56 ->
# 624 "cil/src/frontc/clexer.mll"
               (dbgToken (LBRACE (currentLoc ())))
# 2213 "cil/src/frontc/clexer.ml"

  | 57 ->
# 625 "cil/src/frontc/clexer.mll"
               (dbgToken (RBRACE (currentLoc ())))
# 2218 "cil/src/frontc/clexer.ml"

  | 58 ->
# 626 "cil/src/frontc/clexer.mll"
          (LBRACKET)
# 2223 "cil/src/frontc/clexer.ml"

  | 59 ->
# 627 "cil/src/frontc/clexer.mll"
          (RBRACKET)
# 2228 "cil/src/frontc/clexer.ml"

  | 60 ->
# 628 "cil/src/frontc/clexer.mll"
               (dbgToken (LPAREN (currentLoc ())) )
# 2233 "cil/src/frontc/clexer.ml"

  | 61 ->
# 629 "cil/src/frontc/clexer.mll"
          (RPAREN)
# 2238 "cil/src/frontc/clexer.ml"

  | 62 ->
# 630 "cil/src/frontc/clexer.mll"
               (dbgToken (SEMICOLON (currentLoc ())) )
# 2243 "cil/src/frontc/clexer.ml"

  | 63 ->
# 631 "cil/src/frontc/clexer.mll"
          (COMMA)
# 2248 "cil/src/frontc/clexer.ml"

  | 64 ->
# 632 "cil/src/frontc/clexer.mll"
          (DOT)
# 2253 "cil/src/frontc/clexer.ml"

  | 65 ->
# 633 "cil/src/frontc/clexer.mll"
             (SIZEOF (currentLoc ()))
# 2258 "cil/src/frontc/clexer.ml"

  | 66 ->
# 634 "cil/src/frontc/clexer.mll"
                                        ( if !Cprint.msvcMode then
                                             MSASM (msasm lexbuf, currentLoc ())
                                          else (ASM (currentLoc ())) )
# 2265 "cil/src/frontc/clexer.ml"

  | 67 ->
# 639 "cil/src/frontc/clexer.mll"
                                        ( matchingParsOpen := 0;
                                          let _ = matchingpars lexbuf in
                                          initial lexbuf
                                        )
# 2273 "cil/src/frontc/clexer.ml"

  | 68 ->
# 646 "cil/src/frontc/clexer.mll"
                                        (initial lexbuf )
# 2278 "cil/src/frontc/clexer.ml"

  | 69 ->
# 647 "cil/src/frontc/clexer.mll"
           (scan_ident (Lexing.lexeme lexbuf))
# 2283 "cil/src/frontc/clexer.ml"

  | 70 ->
# 649 "cil/src/frontc/clexer.mll"
  ( if is_oneline_ghost() then begin
      exit_oneline_ghost (); RGHOST
    end
    else EOF
  )
# 2292 "cil/src/frontc/clexer.ml"

  | 71 ->
# 654 "cil/src/frontc/clexer.mll"
       (E.parse_error "Invalid symbol")
# 2297 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_initial_rec lexbuf __ocaml_lex_state

and might_end_ghost lexbuf =
    __ocaml_lex_might_end_ghost_rec lexbuf 168
and __ocaml_lex_might_end_ghost_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 657 "cil/src/frontc/clexer.mll"
        ( exit_ghost_code(); RGHOST )
# 2309 "cil/src/frontc/clexer.ml"

  | 1 ->
# 658 "cil/src/frontc/clexer.mll"
       ( STAR (currentLoc()) )
# 2314 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_might_end_ghost_rec lexbuf __ocaml_lex_state

and comment buffer lexbuf =
    __ocaml_lex_comment_rec buffer lexbuf 170
and __ocaml_lex_comment_rec buffer lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 661 "cil/src/frontc/clexer.mll"
                (  )
# 2326 "cil/src/frontc/clexer.ml"

  | 1 ->
# 662 "cil/src/frontc/clexer.mll"
                ( E.parse_error "Unterminated C comment" )
# 2331 "cil/src/frontc/clexer.ml"

  | 2 ->
# 663 "cil/src/frontc/clexer.mll"
                ( lex_comment comment buffer lexbuf )
# 2336 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comment_rec buffer lexbuf __ocaml_lex_state

and onelinecomment buffer lexbuf =
    __ocaml_lex_onelinecomment_rec buffer lexbuf 175
and __ocaml_lex_onelinecomment_rec buffer lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 666 "cil/src/frontc/clexer.mll"
                (  )
# 2348 "cil/src/frontc/clexer.ml"

  | 1 ->
# 667 "cil/src/frontc/clexer.mll"
                ( lex_comment onelinecomment buffer lexbuf )
# 2353 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_onelinecomment_rec buffer lexbuf __ocaml_lex_state

and matchingpars lexbuf =
    __ocaml_lex_matchingpars_rec lexbuf 178
and __ocaml_lex_matchingpars_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 670 "cil/src/frontc/clexer.mll"
                ( E.newline (); matchingpars lexbuf )
# 2365 "cil/src/frontc/clexer.ml"

  | 1 ->
# 671 "cil/src/frontc/clexer.mll"
                ( matchingpars lexbuf )
# 2370 "cil/src/frontc/clexer.ml"

  | 2 ->
# 672 "cil/src/frontc/clexer.mll"
                ( incr matchingParsOpen; matchingpars lexbuf )
# 2375 "cil/src/frontc/clexer.ml"

  | 3 ->
# 673 "cil/src/frontc/clexer.mll"
                ( decr matchingParsOpen;
                  if !matchingParsOpen = 0 then
                     ()
                  else
                     matchingpars lexbuf
                )
# 2385 "cil/src/frontc/clexer.ml"

  | 4 ->
# 679 "cil/src/frontc/clexer.mll"
         ( do_lex_comment comment lexbuf ;
                  matchingpars lexbuf )
# 2391 "cil/src/frontc/clexer.ml"

  | 5 ->
# 681 "cil/src/frontc/clexer.mll"
        ( let _ = str lexbuf in
                  matchingpars lexbuf )
# 2397 "cil/src/frontc/clexer.ml"

  | 6 ->
# 683 "cil/src/frontc/clexer.mll"
                 ( matchingpars lexbuf )
# 2402 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_matchingpars_rec lexbuf __ocaml_lex_state

and hash lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1) ;   __ocaml_lex_hash_rec lexbuf 187
and __ocaml_lex_hash_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 687 "cil/src/frontc/clexer.mll"
        ( E.newline (); initial lexbuf)
# 2414 "cil/src/frontc/clexer.ml"

  | 1 ->
# 688 "cil/src/frontc/clexer.mll"
         ( hash lexbuf)
# 2419 "cil/src/frontc/clexer.ml"

  | 2 ->
# 689 "cil/src/frontc/clexer.mll"
         ( (* We are seeing a line number. This is the number for the
                   * next line *)
                 let s = Lexing.lexeme lexbuf in
                 let lineno = try
                   int_of_string s
                 with Failure ("int_of_string") ->
                   (* the int is too big. *)
                   Kernel.warning "Bad line number in preprocessed file: %s" s;
                   (-1)
                 in
                 E.setCurrentLine (lineno - 1);
                  (* A file name may follow *)
		  file lexbuf )
# 2436 "cil/src/frontc/clexer.ml"

  | 3 ->
# 702 "cil/src/frontc/clexer.mll"
                ( hash lexbuf )
# 2441 "cil/src/frontc/clexer.ml"

  | 4 ->
let
# 705 "cil/src/frontc/clexer.mll"
                                     pragmaName
# 2447 "cil/src/frontc/clexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 706 "cil/src/frontc/clexer.mll"
                ( let here = currentLoc () in
                  PRAGMA_LINE (pragmaName ^ pragma lexbuf, here)
                )
# 2453 "cil/src/frontc/clexer.ml"

  | 5 ->
# 709 "cil/src/frontc/clexer.mll"
                ( pragmaLine := true; PRAGMA (currentLoc ()) )
# 2458 "cil/src/frontc/clexer.ml"

  | 6 ->
# 710 "cil/src/frontc/clexer.mll"
            ( endline lexbuf)
# 2463 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_hash_rec lexbuf __ocaml_lex_state

and file lexbuf =
    __ocaml_lex_file_rec lexbuf 294
and __ocaml_lex_file_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 713 "cil/src/frontc/clexer.mll"
                      (E.newline (); initial lexbuf)
# 2475 "cil/src/frontc/clexer.ml"

  | 1 ->
# 714 "cil/src/frontc/clexer.mll"
          (file lexbuf)
# 2480 "cil/src/frontc/clexer.ml"

  | 2 ->
# 717 "cil/src/frontc/clexer.mll"
                                             (
                                 let n = Lexing.lexeme lexbuf in
				 let n1 = String.sub n 1 ((String.length n) - 4) in
                                 E.setCurrentWorkingDirectory n1;
                                 endline lexbuf )
# 2489 "cil/src/frontc/clexer.ml"

  | 3 ->
# 722 "cil/src/frontc/clexer.mll"
                                (  (* '"' *)
                                   let n = Lexing.lexeme lexbuf in
                                   let n1 = String.sub n 1
                                       ((String.length n) - 2) in
                                   E.setCurrentFile n1;
				 endline lexbuf)
# 2499 "cil/src/frontc/clexer.ml"

  | 4 ->
# 729 "cil/src/frontc/clexer.mll"
      (endline lexbuf)
# 2504 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_file_rec lexbuf __ocaml_lex_state

and endline lexbuf =
    __ocaml_lex_endline_rec lexbuf 304
and __ocaml_lex_endline_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 732 "cil/src/frontc/clexer.mll"
                ( E.newline (); initial lexbuf)
# 2516 "cil/src/frontc/clexer.ml"

  | 1 ->
# 733 "cil/src/frontc/clexer.mll"
                                ( EOF )
# 2521 "cil/src/frontc/clexer.ml"

  | 2 ->
# 734 "cil/src/frontc/clexer.mll"
      ( endline lexbuf)
# 2526 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_endline_rec lexbuf __ocaml_lex_state

and pragma lexbuf =
    __ocaml_lex_pragma_rec lexbuf 308
and __ocaml_lex_pragma_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 737 "cil/src/frontc/clexer.mll"
                        ( E.newline (); "" )
# 2538 "cil/src/frontc/clexer.ml"

  | 1 ->
# 738 "cil/src/frontc/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (pragma lexbuf) )
# 2544 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_pragma_rec lexbuf __ocaml_lex_state

and str lexbuf =
    __ocaml_lex_str_rec lexbuf 311
and __ocaml_lex_str_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 742 "cil/src/frontc/clexer.mll"
                        ([])
# 2556 "cil/src/frontc/clexer.ml"

  | 1 ->
# 743 "cil/src/frontc/clexer.mll"
             (lex_hex_escape str lexbuf)
# 2561 "cil/src/frontc/clexer.ml"

  | 2 ->
# 744 "cil/src/frontc/clexer.mll"
             (lex_oct_escape str lexbuf)
# 2566 "cil/src/frontc/clexer.ml"

  | 3 ->
# 745 "cil/src/frontc/clexer.mll"
          (lex_simple_escape str lexbuf)
# 2571 "cil/src/frontc/clexer.ml"

  | 4 ->
# 746 "cil/src/frontc/clexer.mll"
                        (E.parse_error "unterminated string" )
# 2576 "cil/src/frontc/clexer.ml"

  | 5 ->
# 747 "cil/src/frontc/clexer.mll"
     (lex_unescaped str lexbuf)
# 2581 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_str_rec lexbuf __ocaml_lex_state

and chr lexbuf =
    __ocaml_lex_chr_rec lexbuf 322
and __ocaml_lex_chr_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 750 "cil/src/frontc/clexer.mll"
              ([])
# 2593 "cil/src/frontc/clexer.ml"

  | 1 ->
# 751 "cil/src/frontc/clexer.mll"
             (lex_hex_escape chr lexbuf)
# 2598 "cil/src/frontc/clexer.ml"

  | 2 ->
# 752 "cil/src/frontc/clexer.mll"
             (lex_oct_escape chr lexbuf)
# 2603 "cil/src/frontc/clexer.ml"

  | 3 ->
# 753 "cil/src/frontc/clexer.mll"
          (lex_simple_escape chr lexbuf)
# 2608 "cil/src/frontc/clexer.ml"

  | 4 ->
# 754 "cil/src/frontc/clexer.mll"
                        ( E.parse_error "unterminated char" )
# 2613 "cil/src/frontc/clexer.ml"

  | 5 ->
# 755 "cil/src/frontc/clexer.mll"
     (lex_unescaped chr lexbuf)
# 2618 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_chr_rec lexbuf __ocaml_lex_state

and msasm lexbuf =
    __ocaml_lex_msasm_rec lexbuf 333
and __ocaml_lex_msasm_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 758 "cil/src/frontc/clexer.mll"
                        ( msasm lexbuf )
# 2630 "cil/src/frontc/clexer.ml"

  | 1 ->
# 759 "cil/src/frontc/clexer.mll"
                        ( msasminbrace lexbuf )
# 2635 "cil/src/frontc/clexer.ml"

  | 2 ->
# 760 "cil/src/frontc/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (msasmnobrace lexbuf) )
# 2641 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasm_rec lexbuf __ocaml_lex_state

and msasminbrace lexbuf =
    __ocaml_lex_msasminbrace_rec lexbuf 337
and __ocaml_lex_msasminbrace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 764 "cil/src/frontc/clexer.mll"
                        ( "" )
# 2653 "cil/src/frontc/clexer.ml"

  | 1 ->
# 765 "cil/src/frontc/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (msasminbrace lexbuf) )
# 2659 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasminbrace_rec lexbuf __ocaml_lex_state

and msasmnobrace lexbuf =
    __ocaml_lex_msasmnobrace_rec lexbuf 340
and __ocaml_lex_msasmnobrace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 768 "cil/src/frontc/clexer.mll"
                        ( lexbuf.Lexing.lex_curr_pos <-
                               lexbuf.Lexing.lex_curr_pos - 1;
                          "" )
# 2673 "cil/src/frontc/clexer.ml"

  | 1 ->
# 771 "cil/src/frontc/clexer.mll"
                        ( lexbuf.Lexing.lex_curr_pos <-
                               lexbuf.Lexing.lex_curr_pos - 5;
                          "" )
# 2680 "cil/src/frontc/clexer.ml"

  | 2 ->
# 774 "cil/src/frontc/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in

                          cur ^ (msasmnobrace lexbuf) )
# 2687 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasmnobrace_rec lexbuf __ocaml_lex_state

and annot_first_token lexbuf =
    __ocaml_lex_annot_first_token_rec lexbuf 348
and __ocaml_lex_annot_first_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 779 "cil/src/frontc/clexer.mll"
            (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      Buffer.clear buf;
      enter_ghost_code ();
      LGHOST
    )
# 2704 "cil/src/frontc/clexer.ml"

  | 1 ->
let
# 785 "cil/src/frontc/clexer.mll"
                         c
# 2710 "cil/src/frontc/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 785 "cil/src/frontc/clexer.mll"
                           ( Buffer.add_char buf c; annot_first_token lexbuf )
# 2714 "cil/src/frontc/clexer.ml"

  | 2 ->
# 786 "cil/src/frontc/clexer.mll"
         ( E.newline(); Buffer.add_char buf '\n'; annot_first_token lexbuf )
# 2719 "cil/src/frontc/clexer.ml"

  | 3 ->
# 787 "cil/src/frontc/clexer.mll"
       ( annot_token lexbuf )
# 2724 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_first_token_rec lexbuf __ocaml_lex_state

and annot_token lexbuf =
    __ocaml_lex_annot_token_rec lexbuf 356
and __ocaml_lex_annot_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 789 "cil/src/frontc/clexer.mll"
         ( let s = Buffer.contents buf in
           make_annot ~one_line:false lexbuf s )
# 2737 "cil/src/frontc/clexer.ml"

  | 1 ->
# 791 "cil/src/frontc/clexer.mll"
         ( E.parse_error "Unterminated annotation" )
# 2742 "cil/src/frontc/clexer.ml"

  | 2 ->
# 792 "cil/src/frontc/clexer.mll"
         (E.newline(); Buffer.add_char buf '\n'; annot_token lexbuf )
# 2747 "cil/src/frontc/clexer.ml"

  | 3 ->
let
# 793 "cil/src/frontc/clexer.mll"
         c
# 2753 "cil/src/frontc/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 793 "cil/src/frontc/clexer.mll"
           ( Buffer.add_char buf c; annot_token lexbuf )
# 2757 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_token_rec lexbuf __ocaml_lex_state

and annot_one_line lexbuf =
    __ocaml_lex_annot_one_line_rec lexbuf 362
and __ocaml_lex_annot_one_line_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 796 "cil/src/frontc/clexer.mll"
            (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      enter_oneline_ghost (); LGHOST
    )
# 2772 "cil/src/frontc/clexer.ml"

  | 1 ->
let
# 800 "cil/src/frontc/clexer.mll"
                         c
# 2778 "cil/src/frontc/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 800 "cil/src/frontc/clexer.mll"
                           ( Buffer.add_char buf c; annot_one_line lexbuf )
# 2782 "cil/src/frontc/clexer.ml"

  | 2 ->
# 801 "cil/src/frontc/clexer.mll"
       ( annot_one_line_logic lexbuf )
# 2787 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_one_line_rec lexbuf __ocaml_lex_state

and annot_one_line_logic lexbuf =
    __ocaml_lex_annot_one_line_logic_rec lexbuf 369
and __ocaml_lex_annot_one_line_logic_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 803 "cil/src/frontc/clexer.mll"
         ( make_annot ~one_line:true lexbuf (Buffer.contents buf) )
# 2799 "cil/src/frontc/clexer.ml"

  | 1 ->
let
# 804 "cil/src/frontc/clexer.mll"
         c
# 2805 "cil/src/frontc/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 804 "cil/src/frontc/clexer.mll"
           ( Buffer.add_char buf c; annot_one_line_logic lexbuf )
# 2809 "cil/src/frontc/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_one_line_logic_rec lexbuf __ocaml_lex_state

;;

# 806 "cil/src/frontc/clexer.mll"
 


# 2820 "cil/src/frontc/clexer.ml"
